package org.apidesign.apifest08.currency;

import java.math.BigDecimal;
import java.math.RoundingMode;
import java.util.Currency;

/**
 * Intern implementation of Bidirectional Convertor - this class is only for internal purposes only.
 * This implementation uses BigDecimals to count result value
 * Scale is set by ISO 4217 as default
 *
 * @author Ladislav Vitasek
 */
class CurrencyConvertorImpl extends AbstractCurrencyConvertor {

    protected RoundingMode roundingMode;
    protected int currency1Scale;
    protected int currency2Scale;

    /**
     * Constructor
     * By default it creates convertor with conversion value 1
     *
     * @param currency1       currency for conversion
     * @param currency2       currency for conversion
     * @param provider constants provider
     * @param roundingMode    math rounding mode
     */
    CurrencyConvertorImpl(Currency currency1, Currency currency2, ConversionRatioProvider provider, RoundingMode roundingMode) {
        this(currency1, currency2, provider, roundingMode, currency1.getDefaultFractionDigits(), currency2.getDefaultFractionDigits());
    }

    /**
     * Constructor
     *
     * @param currency1       currency for conversion
     * @param currency2       currency for conversion
     * @param conversionRatioProvider conversion constants provider
     * @param roundingMode    math rounding mode
     * @param conversionRatioProvider constants provider
     * @param currency2Scale
     */
    CurrencyConvertorImpl(Currency currency1, Currency currency2, ConversionRatioProvider conversionRatioProvider, RoundingMode roundingMode, int currency1Scale, int currency2Scale) {
        super(currency1, currency2, conversionRatioProvider);
        this.roundingMode = roundingMode;
        this.currency1Scale = currency1Scale;
        this.currency2Scale = currency2Scale;
    }


    public BigDecimal convertCurrency1ToCurrency2(BigDecimal amountOfMoney) throws CannotConvertException {
        try {
            final BigDecimal result = amountOfMoney.multiply(getConversionRatioProvider().getCurrency1ToCurrency2Constant());
            return result.setScale(getCurrency2Scale(), getRoundingMode());
        } catch (CannotProvideValueException e) {
            throw new CannotConvertException(e);            
        } catch (ArithmeticException e) {
            throw new CannotConvertException(e);
        }
    }


    public BigDecimal convertCurrency2ToCurrency1(BigDecimal amountOfMoney) throws CannotConvertException {
        try {
            final BigDecimal result = amountOfMoney.multiply(getConversionRatioProvider().getCurrency2ToCurrency1Constant());
            return result.setScale(getCurrency1Scale(), getRoundingMode());
        } catch (CannotProvideValueException e) {
            throw new CannotConvertException(e);
        } catch (ArithmeticException e) {
            throw new CannotConvertException(e);
        }
    }
  
    public String toString() {
        return "CurrencyConvertorImpl{" +
                "currency1=" + getCurrency1() +
                ",currency2=" + getCurrency2() +
                ", roundingMode=" + roundingMode +
                ", currency1Scale=" + currency1Scale +
                ", currency2Scale=" + currency2Scale +
                '}';
    }

    RoundingMode getRoundingMode() {
        return roundingMode;
    }    

    int getCurrency1Scale() {
        return currency1Scale;
    }

    int getCurrency2Scale() {
        return currency2Scale;
    }

    void setRoundingMode(RoundingMode roundingMode) {
        this.roundingMode = roundingMode;
    }

    void setCurrency1Scale(int currency1Scale) {
        this.currency1Scale = currency1Scale;
    }

    void setCurrency2Scale(int currency2Scale) {
        this.currency2Scale = currency2Scale;
    }

}
