package org.apidesign.apifest08.currency;


import java.math.BigDecimal;
import java.util.Currency;
import java.util.Set;


/**
 * Convert between two currencies.
 *
 * @author D'Arcy Smith
 * @version 1.1
 */
class ConvertorImpl
    implements ExchangeRateConvertor
{
    /**
     */
    private final ExchangeRate rate;

    /**
     * Constructs a convertor with the specified currencies.
     * 
     * @param a the currency to convert from.
     * @param aRate the exchage rage between from and to.
     * @param b the currency to convert to.
     * @param bRate the exchage rage between to and from.
     * @throws IllegalArgumentException if either any of the arguments are null or if either rate <= 0.
     */
    public ConvertorImpl(final ExchangeRate r)
    {
        if(r == null)
        {
            throw new IllegalArgumentException("r cannot be null");
        }

        rate = r;
    }
    
    /**
     * Convert an amount from one currency to another.
     * 
     * @param from the currency to convert from.
     * @param to the currency to convert to.
     * @param amount the amount to convert.
     * @return the converted amount.
     * @throws IllegalArgumentException if any of the arguments are null.
     * @throws InvalidConversionException if either from or to are not equal to the currencies passed to the constructor.
     */
    public BigDecimal convert(final Currency   from,
                              final Currency   to,
                              final BigDecimal amount)
        throws InvalidConversionException
    {
        return (rate.convert(from, to, amount));
    }

    /**
     * Check to see if converting between the two currencies is possible.
     * 
     * @param from the currency to convert from.
     * @param to the currency to convert to.
     * @return true if the conversion is possible.
     * @throws IllegalArgumentException if either from or to are null.
     */
    public boolean canConvert(final Currency from, final Currency to)
    {
        return (rate.canConvert(from, to));
    }

    /**
     * Get the currencies that the convertor supports.
     * 
     * @return the supported currencies.
     */
    public Set<Currency> getCurrencies()
    {
        return (rate.getCurrencies());
    }

    /**
     * Get the conversion rate between two currencies.
     * 
     * @param from the currency to convert from.
     * @param to the currency to convert to.
     * @return the conversion rate between the two currencies.
     * @throws InvalidConversionException if canConvert would return false.
     * @throws IllegalArgumentException if either from or to are null.
     */
    public BigDecimal getConversionRate(final Currency from, 
                                        final Currency to)
        throws InvalidConversionException
    {                
        return (rate.getConversionRate(from, to));
    }

    /**
     * Check to see if two ConvertorImpls are equal.
     *
     * @param obj the object to check
     * @return if the ConvertorImpls are not the same (cuyrrencies and rates).
     */
    @Override
    public boolean equals(Object obj)
    {
        if (obj == null)
        {
            return false;
        }
        
        if (getClass() != obj.getClass())
        {
            return false;
        }

        final ConvertorImpl other = (ConvertorImpl) obj;

        return (rate.equals(other.rate));
    }

    /**
     * Get the hashCode of the Convertor.
     *
     * @return the hashCode of the convertor.
     */
    @Override
    public int hashCode()
    {
        return (rate.hashCode());
    }

    /**
     * Get the currencyCode of both currencies.
     *
     * @return the currency codes of both currencies.
     */
    @Override
    public String toString()
    {
        return (rate.getCurrencyA().getCurrencyCode() + " to " + rate.getCurrencyB().getCurrencyCode());
    }

    public ExchangeRate getExchangeRate()
    {
        return (rate);
    }
}
