package org.apidesign.apifest08.test;

import java.math.BigDecimal;
import java.util.Currency;
import junit.framework.TestCase;
import org.apidesign.apifest08.currency.Convertor;
import org.apidesign.apifest08.currency.InvalidConversionException;
import org.apidesign.apifest08.currency.OnlineConvertor;

/** The exchange rates are not always the same. They are changing. Day by day,
 * hour by hour, minute by minute. For every bank it is important to always
 * have the actual exchange rate available in the system. That is why let's
 * create a pluggable convertor that will always have up to date value of its
 * exchange rate.
 * <p>
 * The quest for today is to allow 3rd party developer to write a convertor
 * that adjusts its exchange rate everytime it is queried. This convertor is
 * written by independent vendor, the vendor knows only your Convertor API,
 * he does not know how the whole system looks and how the convertor is supposed
 * to be used.
 */
public class Task3Test 
    extends TestCase
{
    private final static Currency CZK;
    private final static Currency SKK;
    private final static Currency USD;

    static
    {
        CZK = Currency.getInstance("CZK");
        SKK = Currency.getInstance("SKK");
        USD = Currency.getInstance("USD");
    }
    
    public Task3Test(String testName)
    {
        super(testName);
    }

    @Override
    protected void setUp() throws Exception
    {
    }

    @Override
    protected void tearDown() throws Exception
    {
    }

    // Backward compatibly enhance your existing API to support following
    // usecases:
    //


    /** Without knowing anything about the surrounding system, write an
     * implementation of convertor that will return different rates everytime
     * it is queried. Convert USD to CZK and vice versa. Start with the rate of
     * 1USD = 16CZK and adjust it in favor of CZK by 0.01 CZK with every query.
     * As soon as you reach 1USD = 15CZK adjust it by 0.01 CZK in favor of USD
     * until you reach 1USD = 16CZK
     *
     * @return new instance of "online" USD and CZK convertor starting with rate 1USD = 16CZK
     * @throws InvalidConversionException 
     */
    public static Convertor createOnlineCZKUSDConvertor() 
        throws InvalidConversionException
    {
        final Convertor convertor;
        
        convertor = new OnlineConvertor(USD,
                                        CZK,
                                        new TestExchangeRateFinder(new BigDecimal("15.00"), 
                                                                   new BigDecimal("16.00"),
                                                                   new BigDecimal("16.00"),
                                                                   new BigDecimal("0.01"),
                                                                   new BigDecimal("-0.01")));

        // initial rate: 1USD = 16CZK
        // 2nd query 1USD = 15.99CZK
        // 3rd query 1USD = 15.98CZK
        // until 1USD = 15.00CZK
        // then 1USD = 15.01CZK
        // then 1USD = 15.02CZK
        // and so on and on up to 1USD = 16CZK
        // and then another round to 15, etc.
        return convertor;
    }

    public void testFewQueriesForOnlineConvertor() 
        throws InvalidConversionException
    {
        Convertor c = createOnlineCZKUSDConvertor();
        doFewQueriesForOnlineConvertor(c);
    }

    static void doFewQueriesForOnlineConvertor(Convertor c) 
        throws InvalidConversionException            
    {
        BigDecimal amount;

        // convert $5 to CZK using c:
        //assertEquals("Result is 80 CZK");
        amount = c.convert(USD, CZK, new BigDecimal("5.00"));
        assertEquals(new BigDecimal("80.00"), amount);

        // convert $8 to CZK using c:
        //assertEquals("Result is 127.92 CZK");
        amount = c.convert(USD, CZK, new BigDecimal("8.00"));
        assertEquals(new BigDecimal("127.92"), amount);

        // convert $1 to CZK using c:
        //assertEquals("Result is 15.98 CZK");
        amount = c.convert(USD, CZK, new BigDecimal("1.00"));
        assertEquals(new BigDecimal("15.98"), amount);

        // convert 15.97CZK to USD using c:
        //assertEquals("Result is 1$");
        amount = c.convert(CZK, USD, new BigDecimal("15.97"));
        assertEquals(new BigDecimal("1.00"), amount);
    }

    /** Join the convertors and show they behave sane.
     */
    public void testOnlineConvertorComposition() throws Exception {
        BigDecimal amount;
        
        Convertor c = Task2Test.merge(
            createOnlineCZKUSDConvertor(),
            Task1Test.createSKKtoCZK()
        );

        // convert 16CZK to SKK using c:
        // assertEquals("Result is 20 SKK");
        amount = c.convert(CZK, SKK, new BigDecimal("16.00"));
        assertEquals(new BigDecimal("20.00"), amount);
        
        // convert 500SKK to CZK using c:
        // assertEquals("Result is 400 CZK");
        amount = c.convert(SKK, CZK, new BigDecimal("500.00"));
        assertEquals(new BigDecimal("400.00"), amount);

        doFewQueriesForOnlineConvertor(c);
    }
}
