package org.apidesign.apifest08.currency;

import java.util.Date;
import org.apidesign.apifest08.currency.ExchangeRateProvider.ExchangeRateRequest;
import org.apidesign.apifest08.currency.ExchangeRateProvider.ExchangeRateResponse;

/**
 * Exchange rate data source.
 * 
 * @author ked
 */
public final class ExchangeRateDataSource<AmountType, IdentifierType> {

    private final IdentifierType currencyAIdentifier;
    private final IdentifierType currencyBIdentifier;
    private final ExchangeRateProvider<AmountType, IdentifierType> exchangeRateProvider;
    private final Date validFrom;
    private final Date validTill;

    private ExchangeRateDataSource(
            IdentifierType currencyAIdentifier,
            IdentifierType currencyBIdentifier,
            ExchangeRateProvider<AmountType, IdentifierType> exchangeRateProvider,
            Date validFrom,
            Date validTill) {
        if (currencyAIdentifier == null ||
                currencyBIdentifier == null ||
                currencyAIdentifier.equals(currencyBIdentifier)) {
            throw new IllegalArgumentException("Inappropriate exchange rates' identifiers!");
        }
        if (validFrom != null &&
                validTill != null &&
                !validTill.after(validFrom)) {
            throw new IllegalArgumentException("Inappropriate exchange rate validity!");
        }

        this.currencyAIdentifier = currencyAIdentifier;
        this.currencyBIdentifier = currencyBIdentifier;
        this.exchangeRateProvider = exchangeRateProvider;
        this.validFrom = validFrom;
        this.validTill = validTill;
    }

    public IdentifierType getCurrencyAIdentifier() {
        return currencyAIdentifier;
    }

    public IdentifierType getCurrencyBIdentifier() {
        return currencyBIdentifier;
    }

    ExchangeRateProvider<AmountType, IdentifierType> getExchangeRateProvider() {
        return exchangeRateProvider;
    }

    public Date getValidFrom() {
        return validFrom != null ? (Date) validFrom.clone() : null;
    }

    public Date getValidTill() {
        return validTill != null ? (Date) validTill.clone() : null;
    }

    public ExchangeRateValue<AmountType, IdentifierType> getExchangeRate() {
        return getExchangeRate(new Date()); // System.currentTimeMillis()
    }

    public ExchangeRateValue<AmountType, IdentifierType> getExchangeRate(Date instant) {
        ExchangeRateRequest<AmountType, IdentifierType> request =
                new ExchangeRateRequest<AmountType, IdentifierType>();
        ExchangeRateResponse<AmountType, IdentifierType> response =
                new ExchangeRateResponse<AmountType, IdentifierType>();

        request.setCurrencyAIdentifier(currencyAIdentifier);
        request.setCurrencyBIdentifier(currencyBIdentifier);
        request.setInstant(instant);

        exchangeRateProvider.getExchangeRate(request, response);

        ExchangeRateValue<AmountType, IdentifierType> result = response.getExchangeRate();
        if (result.getCurrencyA().getIdentifier().equals(currencyAIdentifier) &&
                result.getCurrencyB().getIdentifier().equals(currencyBIdentifier)) {
            return result;
        } else {
            throw new IllegalStateException("Data source's provider returned inappropriate exchange rate!");
        }
    }

    public static <AmountType, IdentifierType> ExchangeRateDataSource<AmountType, IdentifierType> getExchangeRateDataSource(
            IdentifierType currencyAIdentifier,
            IdentifierType currencyBIdentifier,
            ExchangeRateProvider<AmountType, IdentifierType> exchangeRateProvider) {
        return getExchangeRateDataSource(
                currencyAIdentifier,
                currencyBIdentifier,
                exchangeRateProvider,
                null,
                null);
    }

    public static <AmountType, IdentifierType> ExchangeRateDataSource<AmountType, IdentifierType> getExchangeRateDataSource(
            IdentifierType currencyAIdentifier,
            IdentifierType currencyBIdentifier,
            ExchangeRateProvider<AmountType, IdentifierType> exchangeRateProvider,
            Date validFrom,
            Date validTill) {
        return new ExchangeRateDataSource<AmountType, IdentifierType>(
                currencyAIdentifier,
                currencyBIdentifier,
                exchangeRateProvider,
                validFrom,
                validTill);
    }

    @Override
    public boolean equals(Object obj) {
        if (obj == null) {
            return false;
        }
        if (getClass() != obj.getClass()) {
            return false;
        }
        final ExchangeRateDataSource other =
                (ExchangeRateDataSource) obj;
        if (this.currencyAIdentifier != other.currencyAIdentifier &&
                (this.currencyAIdentifier == null || !this.currencyAIdentifier.equals(other.currencyAIdentifier))) {
            return false;
        }
        if (this.currencyBIdentifier != other.currencyBIdentifier &&
                (this.currencyBIdentifier == null || !this.currencyBIdentifier.equals(other.currencyBIdentifier))) {
            return false;
        }
        if (this.exchangeRateProvider != other.exchangeRateProvider &&
                (this.exchangeRateProvider == null || !this.exchangeRateProvider.equals(other.exchangeRateProvider))) {
            return false;
        }
        if (this.validFrom != other.validFrom && (this.validFrom == null || !this.validFrom.equals(other.validFrom))) {
            return false;
        }
        if (this.validTill != other.validTill && (this.validTill == null || !this.validTill.equals(other.validTill))) {
            return false;
        }
        return true;
    }

    @Override
    public int hashCode() {
        int hash = 7;
        hash = 83 * hash + (this.currencyAIdentifier != null ? this.currencyAIdentifier.hashCode() : 0);
        hash = 83 * hash + (this.currencyBIdentifier != null ? this.currencyBIdentifier.hashCode() : 0);
        hash = 83 * hash + (this.exchangeRateProvider != null ? this.exchangeRateProvider.hashCode() : 0);
        return hash;
    }
}
